@@ -1,3 +1,7 @@
+"""digital flow API"""
+
+__version__ = "6.4.17"
+
 
 from .digiflow_identifier import *
 from .digiflow_io import *
@@ -45,6 +49,6 @@ from .validate import (
     group_can_read,
     group_can_write,
     validate_tiff,
+	validate_xml,
 )
-
 from .common import UNSET_LABEL, XMLNS, FallbackLogger
