@@ -1,10 +1,12 @@
-# -*- coding: utf-8 -*-
+"""I/O related helpers"""
 
+import ast
 import email.utils
 import email.mime.text
 import os
 import shutil
 import smtplib
+import urllib3.exceptions
 
 from pathlib import Path
 
@@ -14,6 +16,22 @@ from lxml import etree as ET
 import digiflow.common as dfc
 import digiflow.digiflow_metadata as df_md
 
+# please pylinter
+# pylint:disable=c-extension-no-member
+
+OAI_KWARG_FGROUP_IMG = "fgroup_images"
+DEFAULT_FGROUP_IMG = "MAX"
+OAI_KWARG_FGROUP_OCR = "fgroup_ocr"
+DEFAULT_FGROUP_OCR = "FULLTEXT"
+OAI_KWARG_POSTFUNC = "post_oai"
+OAI_KWARG_REQUESTS = "request_kwargs"
+REQUESTS_DEFAULT_TIMEOUT = 20
+
+_MIME_TYPE_JPG = "image/jpg"
+_MIME_TYPE_JPEG = "image/jpeg"
+
+_JPG_MIMES = [_MIME_TYPE_JPG, _MIME_TYPE_JPEG]
+
 
 class LoadException(Exception):
     """Load of OAI Data failed"""
@@ -54,19 +72,37 @@ class OAILoader:
         self.base_url = base_url
         self.groups = {}
         self.path_mets = None
-        self.key_images = kwargs['group_images']\
-            if 'group_images' in kwargs else 'MAX'
-        self.post_oai = kwargs['post_oai'] if 'post_oai' in kwargs else None
+        self.key_images = kwargs[OAI_KWARG_FGROUP_IMG] \
+            if OAI_KWARG_FGROUP_IMG in kwargs else DEFAULT_FGROUP_IMG
+        self.key_ocr = kwargs[OAI_KWARG_FGROUP_OCR] \
+            if OAI_KWARG_FGROUP_OCR in kwargs else DEFAULT_FGROUP_OCR
+        self.post_oai = df_md.extract_mets
+        if OAI_KWARG_POSTFUNC in kwargs:
+            self.post_oai = kwargs[OAI_KWARG_POSTFUNC]
         self.groups[self.key_images] = []
-        self.key_ocr = kwargs['group_ocr']\
-            if 'group_ocr' in kwargs else 'FULLTEXT'
-        self.request_kwargs = kwargs['request_kwargs']\
-            if 'request_kwargs' in kwargs else {}
+        self.request_kwargs = self._sanitize_kwargs(kwargs)
         self.groups[self.key_ocr] = []
         self.store = None
 
+    def _sanitize_kwargs(self, in_kwargs):
+        top_dict = {}
+        if OAI_KWARG_REQUESTS in in_kwargs:
+            raw_kwargs = in_kwargs[OAI_KWARG_REQUESTS]
+            if isinstance(raw_kwargs, dict):
+                return raw_kwargs
+            top_tokens = [r.strip() for r in raw_kwargs.split(",")]
+            for t in top_tokens:
+                k, v = t.split("=", maxsplit=1)
+                try:
+                    top_dict[k] = ast.literal_eval(v)
+                except ValueError as val_err:
+                    msg = f"Can't use {top_dict} to load: {val_err.args[0]}"
+                    raise LoadException(msg) from val_err
+        return top_dict
+
     def load(self, record_identifier, local_dst, mets_digital_object_identifier=None,
-             skip_resources=False, force_update=False, metadata_format='mets') -> int:
+             skip_resources=False, force_update=False, metadata_format='mets',
+             use_file_id=False) -> int:
         """
         load metadata from OAI with optional caching in-between
         request additional linked resources if required
@@ -81,8 +117,8 @@ class OAILoader:
         loaded = 0
 
         # sanitize url
-        res_url = "{}?verb=GetRecord&metadataPrefix={}&identifier={}".format(
-            self.base_url, metadata_format, record_identifier)
+        ctx = f"verb=GetRecord&metadataPrefix={metadata_format}&identifier={record_identifier}"
+        res_url = f"{self.base_url}?{ctx}"
         self.path_mets = local_dst
         path_res = self._handle_load(res_url, self.path_mets, self.post_oai, force_update)
         if path_res:
@@ -96,17 +132,20 @@ class OAILoader:
 
         # get linked resources
         for k in self.groups:
-            self.groups[k] = mets_reader.get_filegrp_links(group=k)
+            self.groups[k] = mets_reader.get_filegrp_info(group=k)
 
         # if exist, download them too
         post_func = None
-        for k, linked_res_urls in self.groups.items():
+        for k, file_entries in self.groups.items():
             if k == self.key_ocr:
                 post_func = post_oai_store_ocr
-            for linked_res_url in linked_res_urls:
-                res_val_end = linked_res_url.split('/')[-1]
-                res_val_path = self._calculate_path(k, res_val_end)
-                if self._handle_load(linked_res_url, res_val_path, post_func):
+            for mets_file in file_entries:
+                res_url = mets_file.loc_url
+                url_final_token = res_url.split('/')[-1]
+                if use_file_id:
+                    url_final_token = mets_file.file_id
+                local_path = self._calculate_path(mets_file.file_type, k, url_final_token)
+                if self._handle_load(res_url, local_path, post_func):
                     loaded += 1
         return loaded
 
@@ -117,38 +156,39 @@ class OAILoader:
             if stored_path:
                 if not force_load:
                     return None
-                else:
-                    # force update:
-                    # 1. rename existing data
-                    file_name = os.path.basename(stored_path)
-                    file_dir = os.path.dirname(stored_path)
-                    mets_ctime = str(int(os.stat(stored_path).st_mtime))
-                    bkp_mets = file_name.replace('mets', mets_ctime)
-                    os.rename(stored_path, os.path.join(file_dir, bkp_mets))
-                    # 2. download again anyway
-                    data_path = self.load_resource(res_url, res_path, post_func)
-                    if data_path:
-                        self.store.put(data_path)
-                return res_path
-            else:
+                # force update:
+                # 1. rename existing data
+                file_name = os.path.basename(stored_path)
+                file_dir = os.path.dirname(stored_path)
+                mets_ctime = str(int(os.stat(stored_path).st_mtime))
+                bkp_mets = file_name.replace('mets', mets_ctime)
+                os.rename(stored_path, os.path.join(file_dir, bkp_mets))
+                # 2. download again anyway
                 data_path = self.load_resource(res_url, res_path, post_func)
                 if data_path:
                     self.store.put(data_path)
                 return res_path
-        else:
-            return self.load_resource(res_url, res_path, post_func)
+            data_path = self.load_resource(res_url, res_path, post_func)
+            if data_path:
+                self.store.put(data_path)
+            return res_path
+        return self.load_resource(res_url, res_path, post_func)
 
-    def _calculate_path(self, *args):
+    def _calculate_path(self, mime_type, *args):
         """
-        calculate final path depending on some heuristics which
-        fileGrp has been used - 'MAX' means images, not means 'xml'
+        calculate final path depending on used fileGrp
+        'MAX' means images, not means 'xml'
         """
         res_path = os.path.join(str(self.dir_local), os.sep.join(list(args)))
+        if mime_type in _JPG_MIMES and not res_path.endswith('.jpg'):
+            res_path += ".jpg"
         if '/MAX/' in res_path and not res_path.endswith('.jpg'):
             res_path += '.jpg'
-        elif '/FULLTEXT/' in res_path and not res_path.endswith('.xml'):
+        if "xml" in mime_type and not res_path.endswith('.xml'):
+            res_path += ".xml"
+        if '/FULLTEXT/' in res_path and not res_path.endswith('.xml'):
             res_path += '.xml'
-        return res_path
+        return Path(res_path)
 
     def load_resource(self, url, path_local, post_func):
         """
@@ -164,22 +204,27 @@ class OAILoader:
             if post_func and data:
                 # divide METS from XML (OCR-ALTO)
                 # in rather brute force fashion
-                _snippet = data[:512]
+                data_snippet = data[:512]
                 # propably sanitize data, as it might originate
                 # from test-data or *real* requests
-                if not isinstance(_snippet, str):
-                    _snippet = _snippet.decode('utf-8')
-                if dfc.XMLNS['mets'] in _snippet or dfc.XMLNS['oai'] in _snippet:
+                if isinstance(data_snippet, bytes):
+                    data_snippet = data_snippet.decode('utf-8')
+                if dfc.XMLNS['mets'] in str(data_snippet) or dfc.XMLNS['oai'] in str(data_snippet):
                     data = post_func(self.path_mets, data)
-                elif 'http://www.loc.gov/standards/alto' in _snippet:
+                elif 'http://www.loc.gov/standards/alto' in str(data_snippet):
                     data = post_func(local_path, data)
                 else:
                     raise LoadException(f"Can't handle {content_type} from {url}!")
             return local_path
+        except df_md.DigiflowMetadataException as dfm_exc:
+            raise LoadException(dfm_exc) from dfm_exc
         except LoadException as load_exc:
             raise load_exc
+        except urllib3.exceptions.ReadTimeoutError as read_timeout:
+            msg = f"load {url} read timeout: {read_timeout}"
+            raise LoadException(msg) from read_timeout
         except Exception as exc:
-            msg = "processing '{}': {}".format(url, exc)
+            msg = f"load {url} exception: {exc}"
             raise RuntimeError(msg) from exc
 
 
@@ -189,11 +234,11 @@ class OAIFileSweeper:
         parse *.mets.xml to identify files to be deleted
     """
 
-    def __init__(self, path_store, pattern='mets.xml', filegroups=['MAX', ]):
+    def __init__(self, path_store, pattern='mets.xml', filegroups=None):
         self.work_dir = path_store
         self.pattern = pattern
         self.filegroups = filegroups if isinstance(filegroups, list)\
-            else list(filegroups)
+            else ["MAX"]
 
     def sweep(self):
         """remove OAI-Resources from given dir, if any contained"""
@@ -235,13 +280,12 @@ class OAIFileSweeper:
                                    len(list(_parent.iterdir())) == 0:
                                     _parent.rmdir()
                             except PermissionError:
-                                return 'cannot delete {} (insuff. permission)'\
-                                    .format(pth)
-        return (work_dir, total, "{} Mb".format(size >> 20))
+                                return f"cannot delete {pth} due insuff. perm."
+        return (work_dir, total, f"{(size >> 20)} Mb")
 
     def _get_files(self, mets_xml, filegroup):
         xml_root = ET.parse(str(mets_xml)).getroot()
-        xpath = ".//mets:fileGrp[@USE='{}']/mets:file/mets:FLocat".format(filegroup)
+        xpath = f".//mets:fileGrp[@USE='{filegroup}']/mets:file/mets:FLocat"
         locats = xml_root.findall(xpath, {"mets": "http://www.loc.gov/METS/"})
         links = [xl.get('{http://www.w3.org/1999/xlink}href') for xl in locats]
         return [Path(ln).stem for ln in links]
@@ -337,29 +381,23 @@ def request_resource(url: str, path_local: Path, **kwargs):
     status = 0
     result = None
     try:
-        response = requests.get(url, **kwargs)
+        the_timeout = REQUESTS_DEFAULT_TIMEOUT
+        if "timeout" in kwargs:
+            the_timeout = kwargs["timeout"]
+            del kwargs["timeout"]
+        response = requests.get(url, timeout=the_timeout, **kwargs)
         status = response.status_code
         if status >= 400:
-            _inf = "url '{}' returned '{}'".format(url, status)
+            the_info = f"{url} status {status}"
             if status < 500:
-                raise ClientError(_inf)
-            else:
-                raise ServerError(_inf)
+                raise ClientError(the_info)
+            raise ServerError(the_info)
+        content_type = None
         if status == 200:
             content_type = response.headers['Content-Type']
-
             # textual xml data
             if 'text' in content_type or 'xml' in content_type:
                 result = response.content
-                xml_root = ET.fromstring(result)
-                check_error = xml_root.find('.//error', xml_root.nsmap)
-                if check_error is not None:
-                    msg = "the download of {} fails due to: '{}'".format(
-                        url, check_error.text)
-                    raise LoadException(msg)
-                path_local = _sanitize_local_file_extension(
-                    path_local, content_type)
-
             # catch other content types by MIMI sub_type
             # split "<application|image>/<sub_type>"
             elif content_type.split('/')[-1] in ['jpg', 'jpeg', 'pdf', 'png']:
@@ -368,29 +406,27 @@ def request_resource(url: str, path_local: Path, **kwargs):
                 if not isinstance(path_local, Path):
                     path_local = Path(path_local)
                 path_local.write_bytes(response.content)
-
             # if we went this far, something unexpected has been returned
             else:
-                msg = "download {} with unhandled content-type {}".format(
-                    url, content_type)
+                msg = f"download {url} with unhandled content-type {content_type}"
                 raise ContentException(msg)
         return (path_local, result, content_type)
     except (OSError) as exc:
-        msg = "fail to download '{}' to '{}'".format(url, path_local)
+        msg = f"fail to download {url} to {path_local}"
         raise RuntimeError(msg) from exc
 
 
-def _sanitize_local_file_extension(path_local, content_type):
-    if not isinstance(path_local, str):
-        path_local = str(path_local)
-    if 'xml' in content_type and not path_local.endswith('.xml'):
-        path_local += '.xml'
-    elif 'jpeg' in content_type and not path_local.endswith('.jpg'):
-        path_local += '.jpg'
-    elif 'png' in content_type and not path_local.endswith('.png'):
-        path_local += '.png'
-    elif 'pdf' in content_type and not path_local.endswith('.pdf'):
-        path_local += '.pdf'
+def _sanitize_local_file_extension(path_local: Path, content_type):
+    if not isinstance(path_local, Path):
+        path_local = Path(path_local).absolute()
+    if 'xml' in content_type and path_local.suffix != '.xml':
+        path_local = path_local.with_suffix('.xml')
+    elif 'jpeg' in content_type and path_local.suffix != '.jpg':
+        path_local = path_local.with_suffix('.jpg')
+    elif 'png' in content_type and path_local.suffix != '.png':
+        path_local = path_local.with_suffix('.png')
+    elif 'pdf' in content_type and path_local.suffix != '.pdf':
+        path_local = path_local.with_suffix('.pdf')
     return path_local
 
 
@@ -430,7 +466,7 @@ def post_oai_store_ocr(path_local, the_data):
     if isinstance(the_data, str):
         the_data = the_data.encode('utf-8')
     xml_root = ET.fromstring(the_data)
-    df_md.write_xml_file(xml_root, path_local, preamble=None)
+    df_md.write_xml_file(xml_root, path_local, preamble='')
 
 
 def get_enclosed(tokens_str: str, mark_end='}', mark_start='{', func_find='rfind') -> str:
