@@ -4,8 +4,15 @@ from .record_handler import (
     RecordHandlerException,
 )
 from .record_criteria import (
-    Text,
-    State,
     Datetime,
+	Identifier,
+    State,
+    Text,
+)
+from .record_service import (
+    Client, HandlerInformation,
+	RecordRequestHandler,
+	RecordsExhaustedException,
+	RecordsServiceException,
+	run_server
 )
-from .record_service import Client, HandlerInformation, RecordsExhaustedException, run_server
