@@ -1,4 +1,4 @@
-# -*- coding: utf-8 -*-
+"""specification for handling of metadata METS/MODS DFG-flavour"""
 
 import os
 import shutil
@@ -13,13 +13,16 @@ import digiflow as df
 import digiflow.common as dfc
 import digiflow.validate as dfv
 
-from .conftest import TEST_RES, LIB_RES
+from .conftest import TEST_RES, LIB_RES, exchange_textline
+
+# pylint:disable=c-extension-no-member
 
 
+def test_metsreader_kitodo2_volume():
+    """Got correct primary mods for F-stage?
 
-def test_metsreader_kitodo2_mvw():
-    """
-    Correct primary mods?
+    Modified with 6x: as F-stage got single parent
+    further hierarchy contains 1 entry
     """
 
     # arrange
@@ -28,21 +31,36 @@ def test_metsreader_kitodo2_mvw():
     reader = df.MetsReader(path)
 
     # act
-    report = reader.analyze()
+    mets_report: df.MetsReport = reader.report
+    assert mets_report.prime_report
+    prime_report: df.DmdReport = mets_report.prime_report
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'Af'
-    assert 'gvk-ppn' in report.identifiers
-    assert report.identifiers['gvk-ppn'] == '183475917'
-    assert report.hierarchy == [('DMDLOG_0001', 'volume'), ('183475631', 'multivolume_work')]
-    assert report.locations == ['Nr 83 (6)']
+    assert not mets_report.files
+    assert mets_report.hierarchy == [('183475631', 'multivolume_work')]
+    assert prime_report.languages == ['ger']
+    assert prime_report.dmd_type == 'Af'
+    assert prime_report.identifiers
+    assert 'gvk-ppn' in prime_report.identifiers
+    assert prime_report.identifiers['gvk-ppn'] == '183475917'
+    assert prime_report.locations == 'Nr 83 (6)'
+    assert prime_report.licence
+    assert len(prime_report.licence) == 2
+    assert prime_report.licence[0] == ("use and reproduction",
+                                       "http://rightsstatements.org/vocab/InC/1.0/",
+                                       "Urheberrechtsschutz 1.0")
+    assert prime_report.licence[1] == ("out of print work",
+                                       "n.a.",
+                                       "Wahrnehmung der Rechte durch die VG WORT (§ 51 VGG)")
+    assert prime_report.related == ('host', 'gvk-ppn', '183475631')
 
 
 def test_metsreader_report_vd18_cstage():
     """
     Are correct informations extracted for c-stage?
+
+    Modified with 6x: as C-stage has no parents
+    further hierarchy is empty
     """
 
     # arrange
@@ -51,20 +69,30 @@ def test_metsreader_report_vd18_cstage():
     reader = df.MetsReader(path, 'md9427342')
 
     # act
-    report = reader.analyze()
+    mets_report: df.MetsReport = reader.report
+    assert mets_report.prime_report
+    prime_report: df.DmdReport = mets_report.prime_report
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'Ac'
-    assert report.identifiers['ulbhalvd18'] == '211999504'
-    assert report.hierarchy == [('9427342', 'multivolume_work')]
-    assert not report.locations
+    assert not mets_report.files
+    assert mets_report.hierarchy == []
+    assert prime_report.dmd_type == 'Ac'
+    assert prime_report.languages == ['ger']
+    assert prime_report.identifiers
+    assert prime_report.identifiers['ulbhalvd18'] == '211999504'
+    assert not prime_report.locations
 
 
 def test_metsreader_report_vd18_fstage():
     """
     Correct informations extracted for f-stage?
+    Please note:
+        this shelf is located somewhat different =>
+        mods:location/mods:holdinSimple/mods:copyInformation/mods:shelfLocator
+        This one looks evil, since a shelfLocator is *no* hint for copying ;)
+
+    Modified with 6x: as F-stage got single parent
+    further hierarchy contains 1 entry
     """
 
     # arrange
@@ -73,22 +101,49 @@ def test_metsreader_report_vd18_fstage():
     reader = df.MetsReader(path, 'md9427337')
 
     # act
-    report = reader.analyze()
+    mets_report: df.MetsReport = reader.report
+    assert mets_report.prime_report
+    prime_report: df.DmdReport = mets_report.prime_report
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'Af'
-    assert report.identifiers['ulbhalvd18'] == '211999628'
-    assert report.hierarchy == [
-        ('9427337', 'volume'), ('9427342', 'multivolume_work')]
-    assert report.locations == ['Lb 712 a (3,2)']
+    assert not mets_report.files
+    assert mets_report.hierarchy == [('9427342', 'multivolume_work')]
+    assert prime_report.dmd_type == 'Af'
+    assert prime_report.identifiers
+    assert prime_report.identifiers['ulbhalvd18'] == '211999628'
+    assert prime_report.languages == ['ger']
+    assert prime_report.locations == 'Lb 712 a (3,2)'
+
+
+def test_metsreader_report_k2_goobi_mets():
+    """Ensure Kitodo 2 legacy METS to be archived
+    is still properly recognized"""
+
+    # tests/resources/mets/k2_meta_10261.xml
+    path = os.path.join(TEST_RES, "mets", "k2_meta_10261.xml")
+    assert os.path.exists(path)
+    reader = df.MetsReader(path)
+
+    # act
+    report: df.MetsReport = reader.report
+
+    # assert
+    assert report.hierarchy == [('LOG_0002', 'MultiVolumeWork')]
+    assert report.prime_report
+    prime_report: df.DmdReport = report.prime_report
+    assert prime_report.identifiers
+    assert prime_report.identifiers["goobi:CatalogSourceID"] == "153142537"
+    assert prime_report.identifiers["goobi:anchorID"] == "153142340"
+    assert prime_report.identifiers["urn:nbn"] == "urn:nbn:de:gbv:3:1-1192015415-153142537-13"
 
 
 def test_metsreader_report_vd17_fstage_pica_case():
     """
     Correct informations extracted for VD17 F-stage?
     Is the second letter from PICA preserved?
+
+    Modified with 6x: as F-stage got single parent
+    further hierarchy contains 1 entry
     """
 
     # arrange
@@ -99,21 +154,26 @@ def test_metsreader_report_vd17_fstage_pica_case():
     reader = df.MetsReader(path, 'md14591176')
 
     # act
-    report = reader.analyze()
+    report: df.MetsReport = reader.report
+    assert report.prime_report
+    prime_report: df.DmdReport = report.prime_report
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'AF'
-    assert report.identifiers['pon'] == '008499756'
-    assert report.hierarchy == [
-        ('14591176', 'volume'), ('14591136', 'multivolume_work')]
-    assert report.locations == ['TM0904 (5)']
+    assert not report.files
+    assert report.hierarchy == [('14591136', 'multivolume_work')]
+    assert prime_report.dmd_type == 'AF'
+    assert prime_report.identifiers
+    assert prime_report.identifiers['pon'] == '008499756'
+    assert prime_report.languages == ['ger']
+    assert prime_report.locations == 'TM0904 (5)'
 
 
 def test_metsreader_report_hd_monography():
     """
     Correct extractions for hd monography?
+
+    Modified with 6x: Since monograph stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -125,21 +185,28 @@ def test_metsreader_report_hd_monography():
     reader = df.MetsReader(path, 'md10595')
 
     # act
-    report = reader.analyze()
+    report: df.MetsReport = reader.report
+    assert report.prime_report
+    prime_report: df.DmdReport = report.prime_report
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'Aa'
-    assert report.identifiers['ulbhaldod'] == '187143188'
-    assert report.hierarchy == [('10595', 'monograph')]
-    assert len(report.locations) == 2
-    assert report.locations == ['Pon IIg 694, FK', 'Pon IIg 689, 4° (2)']
+    assert not report.files
+    assert report.hierarchy == []
+    assert prime_report.dmd_type == 'Aa'
+    assert prime_report.languages == ['ger']
+    assert prime_report.identifiers
+    assert prime_report.identifiers['ulbhaldod'] == '187143188'
+    assert prime_report.locations
+    assert len(prime_report.locations) == 2
+    assert prime_report.locations == ['Pon IIg 694, FK', 'Pon IIg 689, 4° (2)']
 
 
 def test_metsreader_report_kitodo2_export_monography():
     """
     Correct extractions for hd monography?
+
+    Modified with 6x: Since monograph stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -148,20 +215,23 @@ def test_metsreader_report_kitodo2_export_monography():
     reader = df.MetsReader(path, 'DMDLOG_0000')
 
     # act
-    report = reader.analyze()
+    report: df.MetsReport = reader.report
+    assert report.prime_report
+    prime_report: df.DmdReport = report.prime_report
     # need to set this manually since
     # we do not know the kitodo ID from METS
-    report.system_identifiers[df.MARK_KITODO2] = '1234'
+    report.system_identifier = {"kitodo2": "1234"}
 
     # assert
-    assert report.languages == ['ger']
-    assert not report.images
-    assert report.type == 'Aa'
-    assert report.identifiers['gvk-ppn'] == '147638674'
-    assert len(report.identifiers) == 3
-    assert report.system_identifiers == {'kitodo2' : '1234'}
-    assert report.hierarchy == [('DMDLOG_0000', 'monograph')]
-    assert report.locations == ['Pon Za 5950, QK']
+    assert not report.files
+    assert report.system_identifier == {"kitodo2": "1234"}
+    assert report.hierarchy == []
+    assert prime_report.languages == ['ger']
+    assert prime_report.dmd_type == 'Aa'
+    assert prime_report.identifiers
+    assert prime_report.identifiers['gvk-ppn'] == '147638674'
+    assert len(prime_report.identifiers) == 3
+    assert prime_report.locations == 'Pon Za 5950, QK'
 
 
 def test_metsreader_logical_type_1686755_is_document():
@@ -169,6 +239,9 @@ def test_metsreader_logical_type_1686755_is_document():
     Check digital object 1686755
     * logical type is 'document'
     * no external PICA-type present (=None)
+
+    Modified with 6x: Since document stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -178,8 +251,8 @@ def test_metsreader_logical_type_1686755_is_document():
     mets_reader = df.MetsReader(mets, 'md1686755')
 
     # assert
-    outcome = mets_reader.get_type_and_hierarchy()
-    assert outcome == (None, 'document', [('1686755', 'document')])
+    outcome = mets_reader.inspect_logical_struct()
+    assert outcome == ("document", [])
 
 
 def test_metsreader_type_pica_monography():
@@ -191,7 +264,7 @@ def test_metsreader_type_pica_monography():
     # act
     mets_reader = df.MetsReader(mets, 'md201517')
 
-    assert "Aa" in mets_reader.get_type_and_hierarchy()
+    assert "monograph" in mets_reader.inspect_logical_struct()
 
 
 def test_metsreader_wrong_logical_type():
@@ -199,6 +272,9 @@ def test_metsreader_wrong_logical_type():
     What happens when encountered an invalid c-stage oai-response?
     Extended Test due data errors from OCR-D-Pilotproject
     => logical type monograph should be multivolume
+
+    Modified with 6x: Since monograph stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -210,8 +286,8 @@ def test_metsreader_wrong_logical_type():
     # in this case it should be "multivolume_work"
     # f-stage 424336
     # f-stage 415691
-    outcome = mets_reader.get_type_and_hierarchy()
-    assert outcome == (None, 'monograph', [('416811', 'monograph')])
+    outcome = mets_reader.inspect_logical_struct()
+    assert outcome == ('monograph', [])
 
 
 @pytest.fixture(name="monograph_hd_invalid_physlinks")
@@ -245,19 +321,27 @@ def test_metsreader_report_for_10595(monograph_hd_invalid_physlinks):
     reader = df.MetsReader(monograph_hd_invalid_physlinks, 'md10595')
 
     # act
-    report = reader.analyze()
+    mets_report: df.MetsReport = reader.report
+    assert mets_report.prime_report
+    report: df.DmdReport = mets_report.prime_report
 
     # assert
-    assert 'Aa' in report.type
+    assert report.dmd_type
+    assert 'Aa' in report.dmd_type
     assert report.languages == ['ger']
+    assert report.identifiers
     assert report.identifiers['ulbhaldod'] == '187143188'
-    assert len(report.links) == 26
+    assert mets_report.links
+    assert len(mets_report.links) == 26
 
 
 def test_metsreader_logical_type_is_multivolume():
     """
     Check that a multivolume_work with 4 volumes is
     correct recognized as pica 'Ac' and 'multivolume_work'
+
+    Modified with 6x: Since C-Stage stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -265,18 +349,20 @@ def test_metsreader_logical_type_is_multivolume():
     mets_reader = df.MetsReader(mets, 'md9427342')
 
     # act
-    mets_reader.analyze()
+    _ = mets_reader.report
 
     # assert
-    outcome = mets_reader.get_type_and_hierarchy()
-    assert outcome == ('Ac', 'multivolume_work', [
-                       ('9427342', 'multivolume_work')])
+    outcome = mets_reader.inspect_logical_struct()
+    assert outcome == ('multivolume_work', [])
 
 
 def test_metsreader_logical_type_is_tome():
     """
     Check that a part of a multivolume_work, 9427334,
     is VL 'tome' and pica 'Af'
+
+    Modified with 6x: as F-stage got single parent
+    further hierarchy contains 1 entry
     """
 
     # arrange
@@ -284,18 +370,21 @@ def test_metsreader_logical_type_is_tome():
     mets_reader = df.MetsReader(mets, 'md9427334')
 
     # act
-    mets_reader.analyze()
+    _ = mets_reader.report
 
     # assert
-    outcome = mets_reader.get_type_and_hierarchy()
-    assert outcome == (
-        'Af', 'tome', [('9427334', 'tome'), ('9427342', 'multivolume_work')])
+    outcome = mets_reader.inspect_logical_struct()
+    assert outcome == ('tome', [('9427342', 'multivolume_work')])
 
 
 def test_metsreader_ambigious_recordinfo():
     """
     Handle strange cornercase in HD collection where digital objects
     posses several recordInfo elements but first match is empty
+
+
+    Modified with 6x: Since monograph stands alone,
+    further hierarchy is empty
     """
 
     # arrange
@@ -303,11 +392,11 @@ def test_metsreader_ambigious_recordinfo():
     mets_reader = df.MetsReader(mets, 'md369765')
 
     # act
-    mets_reader.analyze()
+    _ = mets_reader.report
 
     # assert
-    outcome = mets_reader.get_type_and_hierarchy()
-    assert outcome == (None, 'monograph', [('369765', 'monograph')])
+    outcome = mets_reader.inspect_logical_struct()
+    assert outcome == ('monograph', [])
 
 
 def test_metsreader_clear_agents(tmp_path):
@@ -318,7 +407,7 @@ def test_metsreader_clear_agents(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'migration/369765.mets.xml')
-    the_orig = ET.parse(mets)
+    the_orig = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_agents = the_orig.findall('.//mets:agent', dfc.XMLNS)
     assert len(orig_agents) == 4
     dst = tmp_path / '369765.mets.xml'
@@ -330,7 +419,7 @@ def test_metsreader_clear_agents(tmp_path):
     result_path = mets_reader.write('ulb')
 
     # assert
-    the_root = ET.parse(str(result_path))
+    the_root = ET.parse(str(result_path))  # pyright: ignore[reportCallIssue]
     agents = the_root.findall('.//mets:agent', dfc.XMLNS)
     assert len(agents) == 2
     for agent in agents:
@@ -353,7 +442,7 @@ def test_metsreader_enrich_first_agent(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'k3_300896638-18490701.xml')
-    mets_input = ET.parse(mets)
+    mets_input = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_agents = mets_input.findall('.//mets:agent', dfc.XMLNS)
     assert len(orig_agents) == 0
     dst = tmp_path / 'mets.xml'
@@ -366,7 +455,8 @@ def test_metsreader_enrich_first_agent(tmp_path):
 
     # assert
     assert Path(result_path).exists()
-    dfv.validate_xml(ET.parse(result_path).getroot()) # no Exception plz
+    dfv.validate_xml(ET.parse(result_path).getroot())  # pyright: ignore[reportCallIssue]
+
 
 def test_metsreader_enrich_another_agent(tmp_path):
     """
@@ -383,7 +473,7 @@ def test_metsreader_enrich_another_agent(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'k3_300896638-18490701.xml')
-    mets_input = ET.parse(mets)
+    mets_input = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_agents = mets_input.findall('.//mets:agent', dfc.XMLNS)
     assert len(orig_agents) == 0
     dst = tmp_path / 'mets.xml'
@@ -397,7 +487,7 @@ def test_metsreader_enrich_another_agent(tmp_path):
 
     # assert
     assert Path(result_path).exists()
-    dfv.validate_xml(ET.parse(result_path).getroot()) # no Exception plz
+    dfv.validate_xml(ET.parse(result_path).getroot())  # pyright: ignore[reportCallIssue]
 
 
 def test_metsreader_enrich_agent_kwargs(tmp_path):
@@ -415,7 +505,7 @@ def test_metsreader_enrich_agent_kwargs(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'k3_300896638-18490701.xml')
-    mets_input = ET.parse(mets)
+    mets_input = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_agents = mets_input.findall('.//mets:agent', dfc.XMLNS)
     assert len(orig_agents) == 0
     dst = tmp_path / '369765.mets.xml'
@@ -429,8 +519,8 @@ def test_metsreader_enrich_agent_kwargs(tmp_path):
 
     # assert
     assert Path(result_path).exists()
-    result_root = ET.parse(result_path).getroot()
-    dfv.validate_xml(result_root) # no Exception plz
+    result_root = ET.parse(result_path).getroot()  # pyright: ignore[reportCallIssue]
+    dfv.validate_xml(result_root)  # no Exception plz
     assert result_root.xpath('.//mets:agent[@TYPE="INDIVIDUAL"]/mets:name/text()',
                              namespaces=df.XMLNS)[0] == 'Agent Smith'
 
@@ -446,13 +536,13 @@ def test_metsreader_log_hierarchy_menadoc_oai_record_section():
     target_file = os.path.join(TEST_RES, 'migration/20586.mets.xml')
 
     # act
-    _mreader = df.MetsReader(target_file)
-    with pytest.raises(RuntimeError) as rer:
-        _report = _mreader.report
-    assert 'no record _identifiers' in rer.value.args[0]
+    mreader = df.MetsReader(target_file)
+    with pytest.raises(df.DigiflowMetadataException) as rer:
+        _ = mreader.report
+    assert 'no identifiers' in rer.value.args[0]
 
     # assert - no, this is not happening!
-    #assert _report.identifiers['menadoc.bibliothek.uni-halle.de/dmg'] == '20586'
+    # assert _report.identifiers['menadoc.bibliothek.uni-halle.de/dmg'] == '20586'
 
 
 def test_metsreader_opendata2_inspect_migrated_record_identifiers():
@@ -471,19 +561,22 @@ def test_metsreader_opendata2_inspect_migrated_record_identifiers():
     mets_reader = df.MetsReader(target_file)
 
     # act
-    report = mets_reader.analyze()
+    report: df.MetsReport = mets_reader.report
+    assert report.prime_report
+    dmd_report: df.DmdReport = report.prime_report
 
     # assert
     assert mets_reader.dmd_id == 'md998423'
-    assert len(report.identifiers) == 5
-    assert report.identifiers == {
+    assert dmd_report.identifiers
+    assert len(dmd_report.identifiers) == 5
+    assert dmd_report.identifiers == {
         'urn': 'urn:nbn:de:gbv:3:1-507459',
         'vd16': 'ZV 932',
         'bvb': 'VD0034491',
         'gvk-ppn': '567526844',
         'doi': 'doi:10.25673/opendata2-4398'
     }
-    assert report.system_identifiers == {
+    assert report.system_identifier == {
         'legacy vlid': '998423',
         'opendata2.uni-halle.de': 'https://opendata2.uni-halle.de//handle/1516514412012/4400',
     }
@@ -498,19 +591,22 @@ def test_metsreader_opendata_migrated_record_with_doi():
     mets_reader = df.MetsReader(target_file)
 
     # act
-    report = mets_reader.analyze()
+    report: df.MetsReport = mets_reader.report
+    assert report.prime_report
+    dmd_report: df.DmdReport = report.prime_report
 
     # assert
     assert mets_reader.dmd_id == 'md1177525'
-    assert len(report.identifiers) == 5
-    assert report.identifiers == {
+    assert dmd_report.identifiers
+    assert len(dmd_report.identifiers) == 5
+    assert dmd_report.identifiers == {
         'urn': 'urn:nbn:de:gbv:3:1-132151',
         'gvk-ppn': '216311322',
         'doi': 'doi:10.25673/41099',
         'gbv': '216311322',
         'vd18': '10078320',
     }
-    assert report.system_identifiers == {
+    assert report.system_identifier == {
         'legacy vlid': '1177525',
         'opendata.uni-halle.de': 'https://opendata.uni-halle.de//handle/1981185920/43053',
     }
@@ -524,24 +620,104 @@ def test_metsreader_opendata_inspect_migrated_record_origins():
     mets_reader = df.MetsReader(target_file)
 
     # act
-    report = mets_reader.analyze()
+    assert mets_reader.report
+    assert mets_reader.report.prime_report
+    report: df.DmdReport = mets_reader.report.prime_report
+
+    # without using set get 3 origins, which is unwanted
+    assert report.origins
+    assert len(report.origins) == 2
+    assert report.origins == [("publication", "1574", "Freiberg"),
+                              ("digitization", "2013", "Halle (Saale)")]
+
+
+def test_metsreader_opendata_inspect_publication():
+    """Determine behavior with several publication origins"""
+
+    # arrange
+    target_file = os.path.join(TEST_RES, 'k2_mets_vd18_147638674.xml')
+    mets_reader = df.MetsReader(target_file)
+
+    # act
+    assert mets_reader.report
+    assert mets_reader.report.prime_report
+    report: df.DmdReport = mets_reader.report.prime_report
+
+    # what happened to gnd-link?
+    assert report.origins
+    assert len(report.origins) == 2
+    assert report.origins == [("publication", "1720", "[Dresden](http://d-nb.info/gnd/4012995-0)"),
+                              ("digitization", "2021", "Halle (Saale)")]
+
+
+def test_metsreader_opendata_manipulated_origin(tmp_path):
+    """Fix behavior for records with structs like
+           <mods:place>
+                <mods:placeTerm type="text">London</mods:placeTerm>
+            </mods:place>
+            <mods:place>
+                <mods:placeTerm type="code">XA-GB</mods:placeTerm>
+            </mods:place>
+    which before yielded no results at all
+    """
+
+    # arrange
+    target_file = os.path.join(TEST_RES, 'k2_mets_vd18_147638674.xml')
+    tmp_path = shutil.copyfile(target_file, os.path.join(tmp_path, '147638674.xml'))
+    exchange_textline(tmp_path, {
+        '<mods:placeTerm type="text" valueURI="http://d-nb.info/gnd/4012995-0">Dresden</mods:placeTerm>':
+        '<mods:placeTerm type="code">XA-GB</mods:placeTerm>'
+    })
+
+    # act
+    mets_reader = df.MetsReader(tmp_path)
+    assert mets_reader.report
+    assert mets_reader.report.prime_report
+    report: df.DmdReport = mets_reader.report.prime_report
+    assert report.origins
+    assert len(report.origins) == 2
+    assert report.origins == [("publication", "1720", "Dresden"),
+                              ("digitization", "2021", "Halle (Saale)")]
+
+
+# k2_mets_vd18_147638674
+
+def test_metsreader_opendata_inspect_kitodo3_mono_origins():
+    """How to handle latest Kitodo 3 DMS export"""
+
+    # arrange
+    target_file = os.path.join(TEST_RES, "kitodo3", "1906264740.kxp")
+    mets_reader = df.MetsReader(target_file)
+
+    # act
+    assert mets_reader.report
+    assert mets_reader.report.prime_report
+    report: df.DmdReport = mets_reader.report.prime_report
 
     # 3 origins, which is of course wrong
-    assert report.origins == [('n.a.', 'Freiberg'), ('n.a.', 'Freiberg'),
-                              ('digitization', 'Halle (Saale)')]
+    assert report.origins
+    assert len(report.origins) == 2
+    assert report.origins == [("publication", "1560", "[Wien](http://d-nb.info/gnd/4066009-6)"),
+                              ("digitization", "2025", "Halle (Saale)")]
+    assert report.licence == ('use and reproduction',
+                              'https://creativecommons.org/publicdomain/mark/1.0/',
+                              'Public Domain Mark 1.0')
 
 
 def test_metsreader_zd1_issue_16767392():
-    """Check METS-Reader-output"""
+    """Check METS-Reader-output for digital object
+    with logical type issue"""
 
     # arrange
     mets = os.path.join(TEST_RES, 'vls/zd/zd1-16767392.oai.xml')
 
     # act
-    mets_reader = df.MetsReader(mets, 'md16767392')
+    mets_reader: df.MetsReader = df.MetsReader(mets, 'md16767392')
+    mods_reader: df.ModsReader = df.ModsReader(mets_reader.primary_dmd, "md16767392")
 
-    assert "16767392" in mets_reader.get_identifiers()['local']
-    assert "issue" in mets_reader.get_type_and_hierarchy()
+    assert "16767392" in mods_reader.get_identifiers()['local']
+    the_tree = mets_reader.inspect_logical_struct()
+    assert ('issue', [('16602862', 'year'), ('16289662', 'newspaper')]) == the_tree
 
 
 def test_metsreader_zd1_issue_16359609():
@@ -551,10 +727,14 @@ def test_metsreader_zd1_issue_16359609():
     mets = os.path.join(TEST_RES, 'vls/zd/zd1-16359609.mets.xml')
 
     # act
-    mets_reader = df.MetsReader(mets)
+    mets_reader: df.MetsReader = df.MetsReader(mets)
+    assert mets_reader.report
+    assert mets_reader.report.prime_report
+    dmd_report: df.DmdReport = mets_reader.report.prime_report
 
-    assert "ulbhalvd:16359609" == mets_reader.get_identifiers()['local']
-    assert "issue" in mets_reader.get_type_and_hierarchy()
+    assert dmd_report.identifiers
+    assert "ulbhalvd:16359609" == dmd_report.identifiers['local']
+    assert "issue" in mets_reader.inspect_logical_struct()
 
 
 def test_metsprocessor_clear_filegroups_migration_vd17(tmp_path):
@@ -562,7 +742,7 @@ def test_metsprocessor_clear_filegroups_migration_vd17(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'migration/vd17-14591176.mets.xml')
-    the_orig = ET.parse(mets)
+    the_orig = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_file_groups = the_orig.findall('.//mets:fileGrp', dfc.XMLNS)
     assert len(orig_file_groups) == 6
 
@@ -573,7 +753,7 @@ def test_metsprocessor_clear_filegroups_migration_vd17(tmp_path):
     # act
     mets_proc.clear_filegroups(black_list=['TEASER', 'DOWNLOAD', 'DEFAULT', 'THUMBS', 'MIN'])
     mets_proc.write()
-    new_tree = ET.parse(str(dst))
+    new_tree = ET.parse(str(dst))  # pyright: ignore[reportCallIssue]
     assert len(new_tree.findall('.//mets:fileGrp', dfc.XMLNS)) == 1
     dfv.validate_xml(new_tree.getroot())
 
@@ -583,7 +763,7 @@ def test_metsprocessor_clear_filegroups_odem_ocrd(tmp_path):
 
     # arrange
     mets = os.path.join(TEST_RES, 'opendata/1981185920_38841.xml')
-    the_orig = ET.parse(mets)
+    the_orig = ET.parse(mets)  # pyright: ignore[reportCallIssue]
     orig_file_groups = the_orig.findall('.//mets:fileGrp', dfc.XMLNS)
     assert len(orig_file_groups) == 5
 
@@ -594,9 +774,9 @@ def test_metsprocessor_clear_filegroups_odem_ocrd(tmp_path):
     # act
     mets_proc.clear_filegroups(black_list=['DOWNLOAD', 'DEFAULT', 'THUMBS', 'FULLTEXT'])
     mets_proc.write()
-    
+
     # assert
-    new_tree = ET.parse(dst)
+    new_tree = ET.parse(dst)  # pyright: ignore[reportCallIssue]
     dfv.validate_xml(new_tree.getroot())
     assert len(new_tree.findall('.//mets:fileGrp', dfc.XMLNS)) == 1
 
@@ -608,22 +788,22 @@ def test_metsreader_zd2_issue_18680621():
     mets = os.path.join(TEST_RES, 'kitodo3-zd2/1021634069-18680621.xml')
 
     # act
-    mets_reader = df.MetsReader(mets)
+    mets_reader: df.MetsReader = df.MetsReader(mets)
+    mods_reader: df.ModsReader = df.ModsReader(mets_reader.primary_dmd, mets_reader.dmd_id)
 
-    _idents = mets_reader.get_identifiers()
-    assert _idents == {
+    the_idents = mods_reader.get_identifiers()
+    assert the_idents == {
         'urn': 'urn:nbn:de:gbv:3:1-171133730-102163406918680621-11',
         'kxp-ppn': '102163406918680621',
     }
-    assert mets_reader._system_identifiers() == {'kitodo3': '4583'}
-    _pica, _type, _tree = mets_reader.get_type_and_hierarchy()
-    assert _pica == 'AB'
-    assert _type == 'issue'
-    assert _tree == [('uuid-bb586c88-d9b1-3cef-9a91-4e9f4fa5cc8a', 'issue'),
-                     ('n.a.', 'day'),
-                     ('n.a.', 'month'),
-                     ('1021634069_1868', 'year'),
-                     ('1021634069', 'newspaper')]
+    assert mets_reader.ulb_system_identifier() == {'kitodo3': '4583'}
+    log_type, hierarchy = mets_reader.inspect_logical_struct()
+    assert mods_reader.get_type() == 'AB'
+    assert log_type == 'issue'
+    assert hierarchy == [("uuid-b1b11d08-e2d6-45bd-84ef-0b4495013cff", 'day'),
+                         ("uuid-479b0e34-6221-4e61-bfa4-4558a185d1bf", 'month'),
+                         ('1021634069_1868', 'year'),
+                         ('1021634069', 'newspaper')]
 
 
 @pytest.mark.parametrize(["mets_path", "dmd_id"], [
@@ -636,12 +816,13 @@ def test_metsreader_zd2_issue_18680621():
     (os.path.join(TEST_RES, 'opendata/123456789_27949.xml'), 'md1180329'),
 ])
 def test_metsreader_identify_prime_dmd_section(mets_path, dmd_id):
+    """Ensure hit proper primaray MODS DMD section"""
 
     # act
     mets_reader = df.MetsReader(mets_path)
 
     # assert
-    assert mets_reader._prime_mods_id == dmd_id
+    assert mets_reader.dmd_id == dmd_id
 
 
 def test_metsprocessor_remove_elements_and_close_tags(tmp_path):
@@ -660,7 +841,7 @@ def test_metsprocessor_remove_elements_and_close_tags(tmp_path):
     mets_proc = df.MetsProcessor(str(dst))
 
     # act
-    mets_proc.remove(['alto:Shape', 'alto:Processing'])
+    mets_proc.remove(['Shape', 'Processing'])
     path_result = mets_proc.write()
 
     # assert
@@ -671,7 +852,7 @@ def test_metsprocessor_remove_elements_and_close_tags(tmp_path):
 
 def test_metsprocessor_remove_elements_no_keyerror(tmp_path):
     """Ensure legacy namespaces don't cause trouble anymore
-    
+
     Test Target: Prevent regression
     """
 
@@ -683,7 +864,7 @@ def test_metsprocessor_remove_elements_no_keyerror(tmp_path):
     assert len(mets_proc.xpath('//vl:sourceinfo')) == 1
 
     # act
-    mets_proc.remove(['vl:sourceinfo'])
+    mets_proc.remove(['sourceinfo'])
     path_result = mets_proc.write()
 
     # assert
@@ -707,13 +888,12 @@ def test_metsreader_logical_hierachy_newspaper_issue():
     mets_reader = df.MetsReader(mets, 'md16359603')
 
     # act
-    pica, log_typ, hierachy = mets_reader.get_type_and_hierarchy()
+    log_typ, hierachy = mets_reader.inspect_logical_struct()
 
     # assert
-    assert not pica
     assert log_typ == 'issue'
-    assert hierachy[1:] == [('day0322', 'day'), ('month03', 'month'),
-                            ('17308997', 'year'), ('16289661', 'newspaper')]
+    assert hierachy == [('day0322', 'day'), ('month03', 'month'),
+                        ('17308997', 'year'), ('16289661', 'newspaper')]
 
 
 def test_metsreader_missing_struct_mapping():
@@ -731,8 +911,8 @@ def test_metsreader_missing_struct_mapping():
     assert mets_reader.dmd_id == 'md1177525'
 
     # act
-    with pytest.raises(RuntimeError) as _runtime_error:
-        mets_reader.check()
+    with pytest.raises(df.DigiflowMetadataException) as _runtime_error:
+        mets_reader.inspect_logical_struct_links()
 
     # assert
     assert " no link for logical section:'log1646693'" in _runtime_error.value.args[0]
@@ -744,60 +924,63 @@ DIGIFLOW_CONFIG = LIB_RES / 'digilife.ini'
 def test_metsreader3_report_vd18_cstage():
     """
     Are correct informations extracted for c-stage?
+    * no more further METS hierarchy
+    * no related-something in primary MODS
     """
 
     # arrange
     path = os.path.join(TEST_RES, 'migration', '9427342.mets.xml')
     assert os.path.exists(path)
     reader = df.MetsReader(path, 'md9427342')
-    reader.config = DIGIFLOW_CONFIG
-    _identifier_opts = reader.config.options('identifier')
-    assert len(_identifier_opts) == 14
 
     # act
-    _report = reader.report
+    report = reader.report
+    assert report.prime_report
+    dmd_report: df.DmdReport = report.prime_report
 
     # assert
-    assert _report.languages == ['ger']
-    assert not _report.images
-    assert _report.system_identifiers == {'digitale.bibliothek.uni-halle.de/vd18': '9427342'}
-    assert _report.identifiers == {'gbv': '211999504',
-                                   'ulbhalvd18': '211999504',
-                                   'urn': 'urn:nbn:de:gbv:3:1-636051',
-                                   'vd18': '11228628'}
-    assert _report.type == 'Ac'
-    assert _report.hierarchy == [('9427342', 'multivolume_work')]
-    assert not _report.locations
+    assert not report.files
+    assert report.system_identifier == {'digitale.bibliothek.uni-halle.de/vd18': '9427342'}
+    assert report.hierarchy == []
+    assert dmd_report.languages == ['ger']
+    assert dmd_report.identifiers == {'gbv': '211999504',
+                                      'ulbhalvd18': '211999504',
+                                      'urn': 'urn:nbn:de:gbv:3:1-636051',
+                                      'vd18': '11228628'}
+    assert dmd_report.dmd_type == 'Ac'
+    assert not dmd_report.locations
+    assert dmd_report.related == []
 
 
 def test_metsreader3_report_vd18_fstage():
     """
     Correct informations extracted for f-stage?
+
+    Modified with 6x: as F-stage got single parent
+    further hierarchy contains 1 entry
     """
 
     # arrange
     path = os.path.join(TEST_RES, 'migration', '9427337.mets.xml')
     assert os.path.exists(path)
     reader = df.MetsReader(path, 'md9427337')
-    reader.config = DIGIFLOW_CONFIG
 
     # act
-    _report = reader.report
+    mets_report = reader.report
+    assert mets_report.prime_report
+    dmd_report: df.DmdReport = mets_report.prime_report
 
     # assert
-    assert _report.languages == ['ger']
-    assert not _report.images
-    assert _report.type == 'Af'
-    assert _report.system_identifiers == {
-        'digitale.bibliothek.uni-halle.de/vd18': '9427337'  
-    }
-    assert _report.identifiers == {'gbv': '211999628',
-                                   'ulbhalvd18': '211999628',
-                                   'urn': 'urn:nbn:de:gbv:3:1-635986',
-                                   'vd18': '90311817'}
-    assert _report.hierarchy == [
-        ('9427337', 'volume'), ('9427342', 'multivolume_work')]
-    assert _report.locations == ['Lb 712 a (3,2)']
+    assert dmd_report.languages == ['ger']
+    assert not mets_report.files
+    assert mets_report.logical_type == "volume"
+    assert mets_report.hierarchy == [('9427342', 'multivolume_work')]
+    assert mets_report.system_identifier == {'digitale.bibliothek.uni-halle.de/vd18': '9427337'}
+    assert dmd_report.identifiers == {'gbv': '211999628',
+                                      'ulbhalvd18': '211999628',
+                                      'urn': 'urn:nbn:de:gbv:3:1-635986',
+                                      'vd18': '90311817'}
+    assert dmd_report.locations == 'Lb 712 a (3,2)'
 
 
 def test_metsreader_kitodo2_mena_periodical_volume_dmd_id():
@@ -831,3 +1014,111 @@ def test_metadata_processor_contains_multiple_fgroup():
 
     _proc = df.MetsProcessor(TEST_RES / 'k2_mets_morbio_1748529021.xml')
     assert _proc.contains_group(['MAX'])
+
+
+def test_mets_reader_some_sbb_mets():
+    """Explore SBB METS/MODS"""
+
+    the_reader = df.MetsReader(TEST_RES / "mets" / "SBB_PPN1000056597.xml")
+    the_report = the_reader.report
+    assert the_report.prime_report
+    dmd_report: df.DmdReport = the_report.prime_report
+    assert the_report.logical_type == "monograph"
+    assert dmd_report.languages == ["ger"]
+
+
+def test_mets_reader_newspaper_year_1921():
+    """Behavior if year METS encountered"""
+
+    the_reader = df.MetsReader(TEST_RES / "mets" / "newspaper" /
+                               "1516514412012_175735_year_1921.xml")
+    the_report = the_reader.report
+    assert the_report.prime_report
+    dmd_report: df.DmdReport = the_report.prime_report
+    assert the_report.logical_type == "year"
+    assert dmd_report.languages == ["ger"]
+
+
+def test_metsreader_kitodo2_058141367():
+    """Prevent undesired behavior: subsequent
+    calls created ever more dmd_reports!
+    """
+
+    # arrange
+    mets = TEST_RES / "k2-mets-058141367.k2x"
+    assert mets.is_file()
+
+    # act
+    mets_reader_01 = df.MetsReader(mets)
+    assert mets_reader_01.primary_dmd is not None
+    mods_idents = mets_reader_01.primary_dmd.xpath(
+        'mods:recordInfo/mods:recordIdentifier/text()',
+        namespaces=df.XMLNS)
+
+    # assert
+    assert mods_idents[0] == "058141367"
+    assert mets_reader_01.report
+    assert mets_reader_01.report.dmd_reports
+    assert len(mets_reader_01.report.dmd_reports) == 1
+
+    # re-act
+    _ = df.MetsReader(mets).report  # thouse created next dmd_report entry
+    mets_reader_02 = df.MetsReader(mets)
+    assert mets_reader_02.report
+    assert mets_reader_02.report.dmd_reports
+    assert len(mets_reader_02.report.dmd_reports) == 1
+
+
+def test_metsreader_dataclass():
+    """Ensure unset attribute exists and is
+    initialized with 'None' using Python 3.10+
+    and prevent ... object has no attribute ... AttributeErrors
+    """
+
+    report = df.MetsReport()
+    assert report.system_identifier is None
+
+
+TEST_XML_AS_STRING = '''
+<root>
+    <child>test xmlprocessing to initialize object text1</child>
+    <child>test xmlprocessing to initialize object text2</child>
+    <child>test xmlprocessing to initialize object text3</child>
+    <child>test xmlprocessing to initialize object text4</child>
+    <child>test xmlprocessing to initialize object text5</child>
+</root>
+'''
+
+
+@pytest.mark.parametrize("input_xml", [
+    (ET.fromstring('<root><child>text</child></root>')),
+    (TEST_RES / "mets" / "k2_meta_10261.xml"),
+    (TEST_XML_AS_STRING),
+    (bytes(TEST_XML_AS_STRING, encoding='utf-8'))
+])
+def test_xmlprocessor_init_with(input_xml):
+    """Test XMLProcessor initialization with various input types
+    Ensure correct root element is set, which is 'root' in
+    case of the above artificial test items or 'mets' for METS files
+    """
+    # arrange
+    proc = None
+
+    # act
+    proc = df.XMLProcessor(input_xml)
+
+    # assert
+    assert proc
+    assert proc.root
+    assert proc.root.tag in ['{http://www.loc.gov/METS/}mets', 'root']
+
+
+@pytest.mark.parametrize("input_xml", [
+    (Path('/not/existing/file.xml')),
+    (123456),
+    (None)
+])
+def test_xmlprocessor_init_with_invalid_input(input_xml):
+    """Test XMLProcessor raises exception for invalid input data"""
+    with pytest.raises(df.DigiflowMetadataException):
+        df.XMLProcessor(input_xml)
