@@ -6,6 +6,7 @@ import json.decoder
 import os
 import shutil
 import sys
+import typing
 import unittest.mock
 
 from pathlib import Path
@@ -13,15 +14,11 @@ from pathlib import Path
 
 import pytest
 
+import digiflow as df
 import digiflow.record as df_r
 
-from digiflow import (
-    LoadException,
-    get_enclosed,
-    request_resource,
-)
 
-from .conftest import TEST_RES, LEGACY_HEADER_STR, write_datalist
+from .conftest import TEST_RES, LEGACY_HEADER_STR, mock_response, write_datalist
 
 ROOT = Path(__file__).parents[1]
 
@@ -53,7 +50,7 @@ def test_record_local_identifiers(urn, local_identifier):
     assert urn in str(record)
 
 
-def test_invalid_input_data(tmp_path):
+def test_invalid_input_data_01(tmp_path):
     """Invalid input data format raises an exception"""
 
     # arrange
@@ -69,6 +66,27 @@ def test_invalid_input_data(tmp_path):
     assert "invalid fields" in str(exc.value)
 
 
+def test_tricky_info_data_01(tmp_path):
+    """Info is partially invalid dictionary, i.e. not parseable by Python's ast
+    Encountered at migration QA when deleting resources"""
+
+    # arrange
+    invalid_path_dir = tmp_path / 'invalid_data'
+    invalid_path_dir.mkdir()
+    invalid_path = invalid_path_dir / 'invalid.tsv'
+    data = ["123\t{\"issue\":\"148 (29.6.1898)\"}\tn.a.\tn.a.\n",
+            "124\t{\"issue\": \"[149] (29.6.1898) Festzeitung des \"General-Anzeiger\" zur 200jährigen Jubelfeier der Francke'schen Stiftungen zu Halle a. S./16559052\"}\tn.a.\tn.a.\n"]
+    write_datalist(invalid_path, data, headers="IDENTIFIER\tINFO\tSTATE\tSTATE_TIME\n")
+
+    handler = df_r.RecordHandler(invalid_path, data_fields=df_r.RECORD_HEADER)
+
+    r1: typing.Optional[df_r.Record] = handler.next_record(new_state="busy")
+    assert r1 is not None
+    assert r1.info["issue"] == "148 (29.6.1898)"
+    r2: df_r.Record = handler.next_record()
+    assert r2.info["issue"] == "[149] (29.6.1898) Festzeitung des \"General-Anzeiger\" zur 200jährigen Jubelfeier der Francke'schen Stiftungen zu Halle a. S./16559052"
+
+
 @pytest.fixture(name="valid_datasets")
 def _fixture_valid_input_data(tmp_path):
     """
@@ -98,14 +116,14 @@ def test_migrate_state_saved(valid_datasets):
 
     # arrange
     handler = df_r.RecordHandler(valid_datasets, mark_lock='ocr_done')
-    a_set = handler.next_record()
+    a_set: df_r.Record = handler.next_record()
     assert a_set.identifier == 'oai:myhost.de/dod:123'
 
     # act
     handler.save_record_state(a_set.identifier, state='ocr_done', INFO='444')
 
     # assert
-    next_dataset = handler.next_record()
+    next_dataset: typing.Optional[df_r.Record] = handler.next_record()
     assert next_dataset.identifier == 'oai:myhost.de/dod:124'
 
     # act
@@ -145,16 +163,18 @@ def test_migration_dataset_vl_formats(vl_datasets):
 
     # arrange
     handler = df_r.RecordHandler(vl_datasets)
-    a_set = handler.next_record()
+    a_set: df_r.Record = handler.next_record()
 
     # assert
-    assert handler.position == "0001/0002"
+    assert a_set.context.position == 1
+    assert a_set.context.total_len == 2
+    assert a_set.context.data_path == vl_datasets
     assert a_set.local_identifier == '1416976'
     assert a_set.set == 'menalib'
     handler.save_record_state(a_set.identifier, 'busy')
 
-    a_set = handler.next_record()
-    assert handler.position == "0002/0002"
+    a_set: df_r.Record = handler.next_record()
+    assert a_set.context.position == 2
     assert a_set.local_identifier == '696'
     assert a_set.set == 'pon##book'
 
@@ -213,7 +233,7 @@ def test_statelist_ocr_hdz(tmp_path):
 
     # act
     record = handler.next_record()
-    assert handler.position == '0001/0001'
+    assert record.context.position == 1
     assert record[df_r.FIELD_IDENTIFIER] == 'oai:digitale.bibliothek.uni-halle.de/zd:123'
     handler.save_record_state(record[df_r.FIELD_IDENTIFIER])
 
@@ -235,7 +255,7 @@ def test_statelist_use_n_a_properly(oai_record_list):
 
     # act
     record = handler.next_record()
-    assert handler.position == '0006/0006'
+    assert record.context.position == 6
     assert record.identifier == 'oai:digitale.bibliothek.uni-halle.de/zd:9510508'
     handler.save_record_state(record.identifier)
 
@@ -496,41 +516,6 @@ def test_records_default_header_from_file(oai_record_list):
     assert frame_handler.states([c_state]) == 2
 
 
-def mock_response(**kwargs):
-    """Create custum mock object"""
-
-    _response = unittest.mock.MagicMock()
-    _response.reason = 'testing reason'
-    if 'reason' in kwargs:
-        _response.reason = kwargs['reason']
-    if 'status_code' in kwargs:
-        _response.status_code = int(kwargs['status_code'])
-    if 'headers' in kwargs:
-        _response.headers = kwargs['headers']
-    if 'data_path' in kwargs:
-        with open(kwargs['data_path'], encoding="utf-8") as xml:
-            _response.content = xml.read().encode()
-    return _response
-
-
-@unittest.mock.patch('requests.get')
-def test_response_200_with_error_content(mock_requests):
-    """test request results into OAILoadException"""
-
-    # arrange
-    data_path = os.path.join(str(ROOT), 'tests/resources/opendata/id_not_exist.xml')
-    _req = mock_response(status_code=200,
-                         headers={'Content-Type': 'text/xml;charset=UTF-8'},
-                         data_path=data_path)
-    mock_requests.return_value = _req
-
-    # act
-    with pytest.raises(LoadException) as exc:
-        request_resource('http://foo.bar', Path())
-
-    assert 'verb requires' in str(exc.value)
-
-
 def test_records_sample_zd1_post_ocr():
     """Ensure proper state recognized"""
 
@@ -652,6 +637,29 @@ def test_record_handler_quotation_mixture_json(tmp_path):
     assert 'Extra data: line 1 column 7 (char 6)' in _decode_err.value.args[0]
 
 
+def test_record_handler_set_next_new_state(tmp_path):
+    """
+    Fix behavior when setting new state for next entry
+    Start at position 3, since first 2 entries already done
+    """
+
+    # arrange
+    path_list_sample = os.path.join(TEST_RES, 'migration-qa-images.csv')
+    path_oai_list1 = tmp_path / 'migration-qa-images.csv'
+    shutil.copy(path_list_sample, path_oai_list1)
+    handler = df_r.RecordHandler(path_oai_list1, transform_func=df_r.row_to_record)
+
+    # act
+    next_01 = handler.next_record(new_state="busy")
+    assert next_01.context.position == 3
+    assert next_01.context.total_len == 16
+    next_02 = handler.next_record(new_state="busy")
+    assert next_02.context.position == 4
+    next_03 = handler.next_record(new_state="busy")
+    assert next_03.context.position == 5
+    assert next_03.context.total_len == 16
+
+
 def test_record_handler_quotation_fixed_json(tmp_path):
     """
     Fix behavior when encountering data which
@@ -670,7 +678,7 @@ def test_record_handler_quotation_fixed_json(tmp_path):
     # act
     _next_rec = handler.next_record(state='ocr_fail')
     assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
-    _info_token = get_enclosed(_next_rec.info)
+    _info_token = df.get_enclosed(_next_rec.info)
     _last_info = json.loads(_info_token)['info']
     assert 'processing https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg' in _last_info
 
@@ -693,7 +701,7 @@ def test_record_handler_quotation_fixed_ast(tmp_path):
     # act
     _next_rec = handler.next_record(state='ocr_fail')
     assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
-    _info_token = get_enclosed(_next_rec.info)
+    _info_token = df.get_enclosed(_next_rec.info)
     _last_info = ast.literal_eval(_info_token)['info']
     assert 'processing https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg' in _last_info
 
@@ -720,7 +728,7 @@ def test_record_handler_quotation_mixture_ast(tmp_path):
     # act
     _next_rec = handler.next_record(state='ocr_fail')
     assert _next_rec.info.startswith('141.48.10.202@2023-01-17_15:55:49')
-    _info_token = get_enclosed(_next_rec.info)
+    _info_token = df.get_enclosed(_next_rec.info)
     _last_info = ast.literal_eval(_info_token)['info']
     assert "processing 'https://opendata.uni-halle.de/retrieve/b7f7f81d-e65f-4c7d-95c6-7384b184c6a9/00001051.jpg'" in _last_info
 
@@ -744,7 +752,7 @@ def test_record_handler_quotation_mixture_ast_639763(tmp_path):
 
     # act
     _next_rec = handler.next_record(state='ocr_busy')
-    _info_token = get_enclosed(_next_rec.info)
+    _info_token = df.get_enclosed(_next_rec.info)
     with pytest.raises(SyntaxError) as _decode_err:
         ast.literal_eval(_info_token)
     assert 'invalid syntax' in _decode_err.value.args[0]
@@ -770,7 +778,7 @@ def test_record_handler_quotation_mixture_ast_639763_python_310(tmp_path):
 
     # act
     _next_rec = handler.next_record(state='ocr_busy')
-    _info_token = get_enclosed(_next_rec.info)
+    _info_token = df.get_enclosed(_next_rec.info)
     with pytest.raises(SyntaxError) as _decode_err:
         ast.literal_eval(_info_token)
     assert "unmatched '}'" in _decode_err.value.args[0]
@@ -778,8 +786,8 @@ def test_record_handler_quotation_mixture_ast_639763_python_310(tmp_path):
 
 def test_record_handler_with_broken_row(tmp_path):
     """
-    Behavior when encountering broken rows
-    Pevent KeyError: 'STATE'
+    Behavior when encountering broken data row No4
+    Prevent KeyError: 'STATE'
     """
 
     # arrange
@@ -796,4 +804,83 @@ def test_record_handler_with_broken_row(tmp_path):
     with pytest.raises(df_r.RecordHandlerException) as handl_exc:
         handler.next_record(state='foo')  # dummy state to provoke error
 
-    assert "line:003 no STATE field " in handl_exc.value.args[0]
+    assert "line:004 no STATE field " in handl_exc.value.args[0]
+
+
+def test_record_handler_merge_info_dicts(tmp_path):
+    """Two lists merged into one and result
+    info field was merged too since it's a literal 
+    python dict.
+
+    Please note:
+    This merge will only work if both INFO fields
+    can evaluate to dictionaries!
+    """
+
+    # arrange
+    path_oai_list_a = tmp_path / 'oai_list_a'
+    data_fresh = [
+        "123\tn.a.\t2015-08-25T20:00:35Z\t{'pages':23, 'ods_created':'1984-10-03'}\tu.a.\tn.a.\n"
+        "124\tn.a.\t2015-08-25T20:00:35Z\t{'pages':24, 'ods_created':'1985-05-05'}\tn.a.\tn.a.\n"
+    ]
+    write_datalist(path_oai_list_a, data_fresh, LEGACY_HEADER_STR)
+    dst_hndlr = df_r.RecordHandler(
+        path_oai_list_a,
+        data_fields=df_r.LEGACY_HEADER,
+        transform_func=df_r.row_to_record)
+
+    list_merge = tmp_path / 'oai_list_b'
+    data2 = [
+        "123\tn.a.\t2015-08-25T20:00:35Z\t{'pages':23, 'n_ocr':20}\tocr_done\t2024-10-18_11:12:00\n",
+    ]
+    write_datalist(list_merge, data2, LEGACY_HEADER_STR)
+
+    # act: must *not* set ignore_state=None,
+    # otherwise first record from second list erases
+    # existing data from list a
+    results = dst_hndlr.merges(list_merge, dry_run=False)
+    merged_record: df_r.Record = dst_hndlr.next_record(state='ocr_done')
+
+    # no merge since state 'n.a.' from list b
+    # first record of list b ignored
+    # to preserve existing data from list a
+    assert results['merges'] == 1
+    assert results['ignores'] == 0
+    assert results['appendeds'] == 0
+    assert dst_hndlr.total_len == 2
+    assert merged_record.info == {'n_ocr': 20, 'pages': 23, 'ods_created': '1984-10-03'}
+
+
+def test_record_handler_merge_write_read(tmp_path):
+    """Two lists info field merged too but managed
+    to handle quotations around the info-string.
+    """
+
+    # arrange
+    path_oai_list_a = tmp_path / 'oai_list_a'
+    data_fresh = [
+        "123\tn.a.\t2015-08-25T20:00:35Z\t{'pages':23, 'ods_created':'1984-10-03'}\tu.a.\tn.a.\n"
+    ]
+    write_datalist(path_oai_list_a, data_fresh, LEGACY_HEADER_STR)
+    dst_hndlr = df_r.RecordHandler(
+        path_oai_list_a,
+        data_fields=df_r.LEGACY_HEADER,
+        transform_func=df_r.row_to_record)
+
+    list_merge = tmp_path / 'oai_list_b'
+    data2 = [
+        "123\tn.a.\t2015-08-25T20:00:35Z\t\"{'xml_invalid': \"Element 'mods:subtitle': This element is not expected.\"}\"\tocr_done\t2024-10-18_11:12:00\n",
+    ]
+    write_datalist(list_merge, data2, LEGACY_HEADER_STR)
+
+    # act
+    dst_hndlr.merges(list_merge, dry_run=False)
+    new_hndlr = df_r.RecordHandler(path_oai_list_a,
+                                   data_fields=df_r.LEGACY_HEADER,
+                                   transform_func=df_r.row_to_record)
+
+    # assert
+    tha_record: df_r.Record = new_hndlr.next_record(state='ocr_done')
+    assert tha_record.info == {'pages': 23,
+                               'ods_created': '1984-10-03',
+                               'xml_invalid': "Element 'mods:subtitle': This element is not expected."}
